#ifndef SRDTSC_H
#define SRDTSC_H

#ifdef __cplusplus
extern "C" {
#endif

/*
 * если определены HAVE_{CLANG,GCC}_BUILTIN_RDTSC, то есть builtin
 * inline asm для clang/gcc одинаковый
 * на POWER AIX/Linux есть библиотечные функции
 * если определено _WIN32, то есть builtin
 * на Solaris нужен asm
 * если ia32/x86_64 и нет HAVE_{CLANG,GCC}_BUILTIN_RDTSC, то нужен asm
 */

#if defined HAVE_GCC_BUILTIN_RDTSC && !(defined ANDROID)
#if defined(PROCESSOR_TYPE) && (PROCESSOR_TYPE == PROC_TYPE_E2K32 || PROCESSOR_TYPE == PROC_TYPE_E2K64)
#include <x86intrin.h>
#endif
static inline unsigned long long leoh_readtsc(void)
{
    return __builtin_ia32_rdtsc();
}
#elif defined HAVE_CLANG_BUILTIN_RDTSC
#include <x86intrin.h>
static inline unsigned long long leoh_readtsc(void)
{
    return _rdtsc();
}
#elif !defined(IOS) && (defined __clang__ || defined __GNUC__) && (PROCESSOR_TYPE == PROC_TYPE_I386 || PROCESSOR_TYPE == PROC_TYPE_X64)
static inline unsigned long long leoh_readtsc(void)
{
    unsigned int tickl, tickh;
    __asm__ __volatile__("rdtsc":"=a"(tickl),"=d"(tickh));
    return ((unsigned long long)tickh << 32)|tickl;
}
#elif defined (__powerpc__) || defined (__powerpc) || defined (__ppc__)
#include <sys/time.h>
#if defined AIX
/*
 * результат в наносекундах
 * http://publib.boulder.ibm.com/infocenter/pseries/v5r3/index.jsp?topic=/com.ibm.aix.basetechref/doc/basetrf2/read_real_time.htm
 */
static inline unsigned long long leoh_readtsc(void)
{
    timebasestruct_t cur_time;
    const unsigned long long ns_per_sec = 1000000000ULL;
    read_real_time(&cur_time, TIMEBASE_SZ);
    return
	ns_per_sec * (unsigned long long)cur_time.tb_high + cur_time.tb_low;
}
#elif defined LINUX  /* POWER LINUX */
#include <sys/platform/ppc.h>
/*
 * https://www.gnu.org/software/libc/manual/html_node/PowerPC.html
 * возвращает uint64_t --> преобразование не нужно
 */
static inline unsigned long long leoh_readtsc(void)
{
    return __ppc_get_timebase();
}
#else
#error POWER at unkonwn OS detected, cant continue
#endif /* POWER или AIX или Linux. На 2016 год других не было */

#elif defined _WIN32
/*
 * как минимум, начиная с VS 2005 есть intrinsic __rdtsc()
 * https://msdn.microsoft.com/ru-ru/library/twchhe95%28v=vs.80%29.aspx
 */
#pragma warning (push)
#pragma warning (disable:4005)
#include <intrin.h>
#pragma warning (pop)

#pragma intrinsic(__rdtsc)
static __inline unsigned long long leoh_readtsc(void)
{
    return (unsigned long long) __rdtsc();
}
#elif defined __SUNPRO_CC || defined __SUNPRO_C
#  if PROCESSOR_TYPE == PROC_TYPE_I386 || PROCESSOR_TYPE == PROC_TYPE_X64
static inline unsigned long long leoh_readtsc(void)
{
    unsigned int tickl, tickh;
    asm("rdtsc" : "=a" (tickl), "=d" (tickh));
    return ((unsigned long long)tickh << 32)|tickl;
}
#  else
unsigned long long leoh_readtsc(void);
#  endif	/* Solaris + SunStudio */
#elif defined(IOS)
# define NO_LEOH_READTSC 1
#elif (defined(UNIX) && defined(PROCESSOR_TYPE) && (PROCESSOR_TYPE == PROC_TYPE_ARM || PROCESSOR_TYPE == PROC_TYPE_ARM64 || PROCESSOR_TYPE == PROC_TYPE_MIPS32))
# define NO_LEOH_READTSC 1
#else

# if !defined NO_LEOH_READTSC
#   define NO_LEOH_READTSC 1
#   error "Unknown platform: Read TSC implementation not found"
# endif
#endif	/* конец перебора вариантов */

#ifdef __cplusplus
}
#endif

#endif /*SRDTSC_H*/
