/*
* Copyright(C) 2000  
*
*    , 
*    .
*
*        ,
* ,    ,
*     ,
* ,      
*     
*      .
*/
#include "rdr_prj.h"


/*     */
static const TCHAR RDR_DEFAULT_RDR[] = _TEXT("\\local\\KeyDevices\\default");
/*      */
static const TCHAR RDR_DEFAULT_GLOBAL_RDR[] = _TEXT("\\CONFIG\\KeyDevices\\default");

DWORD rdr_context_default_get(
    const TSupSysEList *list,
    TSupSysEContext **context)
{
    DWORD code;
    size_t length = 0;
    TCHAR *name = NULL;
    const TCHAR *paths[2] = { RDR_DEFAULT_RDR, RDR_DEFAULT_GLOBAL_RDR };
    int i;
    TSupSysEFind *ctx;

    SUPSYS_PRE_WRITE_PTRS(context, sizeof(TSupSysEContext*));
    for (i = 0; i < 2; i++)
    {
	code = support_registry_get_string(paths[i], &length, NULL);
	if (code
	    && code != (DWORD)ERROR_FILE_NOT_FOUND
	    && code != (DWORD)ERROR_FILE_INVALID)
	    return code;
	if (code)
	    continue;
	name = malloc((length + 1) * sizeof(TCHAR));
	if (name == NULL)
	    return (DWORD)NTE_NO_MEMORY;
	code = support_registry_get_string(RDR_DEFAULT_RDR, &length, name);
	if (code
	    && code != (DWORD)ERROR_FILE_INVALID
	    && code != (DWORD)ERROR_FILE_NOT_FOUND)
	{
	    free(name);
	    return code;
	}
	if (code)
	{
	    free(name);
	    continue;
	}
	code = supsys_parse_internal_name(list, name, context);
	if (!code)
	{
	    free(name);
	    return ERROR_SUCCESS;
	}
	free(name);
    }
    code = supsys_find_open(list, &ctx);
    if (code)
	return code;
    code = supsys_find_next(ctx, context);
    supsys_find_close(ctx);
    if (code == (DWORD)SCARD_E_READER_UNAVAILABLE)
	return (DWORD)NTE_PROVIDER_DLL_FAIL;
    if (code)
	return code;
    return ERROR_SUCCESS;
}

DWORD rdr_context_default_set(const TSupSysEContext *context)
{
    DWORD code;
    TCHAR *str;
    TCHAR *connect = NULL;
    TSupSysNickname nickname;
    size_t length;

    code = supsys_nickname(context, nickname);
    if (code)
	return code;
    code = supsys_connect_current(context, &length, NULL);
    if (code)
	return code;
    if (length)
    {
	connect = malloc((length + 1) * sizeof(TCHAR));
	if (!connect)
	    return (DWORD)NTE_NO_MEMORY;
	code = supsys_connect_current(context, &length, connect);
	if (code)
	{
	    free(connect);
	    return code;
	}
    }
    code = support_default_internal_register_name(context, &length, NULL,
	connect);
    if (code)
    {
	free(connect);
	return code;
    }
    str = malloc((length + 1) * sizeof(TCHAR));
    if (str == NULL)
    {
	free(connect);
	return (DWORD)NTE_NO_MEMORY;
    }
    code = support_default_internal_register_name(context, &length, str,
	connect);
    free(connect);
    if (code)
    {
	free(str);
	return code;
    }
    code = support_registry_put_string(RDR_DEFAULT_RDR, str);
    free(str);
    return code;
}

DWORD rdr_lock(TSupSysEContext *context, const TProvCallCtx * carrier_ctx)
{
    DWORD code;
    DbTrace(DB_CALL, (FTEXT(db_ctx, "(...)")));
    code = supsys_call(context, READER_FUN_LOCK, (TSupSysInfo*)carrier_ctx);
    return code == (DWORD)ERROR_NOT_SUPPORTED ? ERROR_SUCCESS : code;
}

void rdr_unlock(TSupSysEContext *context)
{
    DbTrace(DB_CALL, (FTEXT(db_ctx, "(...)")));
    supsys_call(context, READER_FUN_UNLOCK, NULL);
}

DWORD rdr_connect_carrier(
    TSupSysEContext *context, unsigned int * applet_count)
{
    TReaderInfoConnectCarrier info;
    DWORD code;
    DbTrace(DB_CALL, (FTEXT(db_ctx, "(...)")));
    SUPSYS_PRE(applet_count);

    info.flags = 0;
    info.media_list = (TSupSysEList*)supsys_owner_context(context);
    info.applet_count = 1; //    ,     READER_FUN_CONNECT_CARRIER    
    code = supsys_call(context, READER_FUN_CONNECT_CARRIER, &info);
    if (code && (code != ERROR_NOT_SUPPORTED))
	return code;

    if (info.applet_count > 1)
	*applet_count = info.applet_count;
    else
	*applet_count = 1;
    return ERROR_SUCCESS;
}


DWORD rdr_disconnect_carrier(
    TSupSysEContext *context)
{
    DbTrace(DB_CALL, (FTEXT(db_ctx, "(...)")));
    return supsys_call(context, READER_FUN_DISCONNECT_CARRIER, NULL);
}

DWORD rdr_local_machine(
    TSupSysEContext *context,
    int local_machine)
{
    DWORD code;
    DbTrace(DB_CALL, (FTEXT(db_ctx, "(...)")));
    code = supsys_call(context, READER_FUN_LOCAL_MACHINE, &local_machine);
    if (code == (DWORD)ERROR_NOT_SUPPORTED)
	code = ERROR_SUCCESS;
    return code;
}

DWORD rdr_select_applet(
    TSupSysEContext *context, unsigned applet_number)
{    
    DWORD code;
    DbTrace(DB_CALL, (FTEXT(db_ctx, "(...)")));
    code = supsys_call(context, READER_FUN_SET_APP, &applet_number);
    if( code == (DWORD)ERROR_NOT_SUPPORTED )
	code = ERROR_SUCCESS;
    return code;
}

DWORD rdr_unique_get(
    const TSupSysEContext *context,
    size_t *length,
    TCHAR *unique) {
    TSupSysInfoText info_unqiue_ptr;
    DWORD code;
    DbTrace(DB_CALL, (FTEXT(db_ctx, "(...)")));
    if (!context || !length)
	return (DWORD)ERROR_INVALID_PARAMETER;

    info_unqiue_ptr.length = *length;
    info_unqiue_ptr.text = unique;
    code = supsys_call(context, READER_FUN_UNIQUE_GET, &info_unqiue_ptr);
    if (code)
	return code;
    *length = info_unqiue_ptr.length;
    return code;
}

DWORD rdr_get_user_defined_name(
    const TSupSysEContext *context,
    size_t *length,
    TCHAR *name) {
    TSupSysInfoText info_name_ptr;
    DWORD code;
    DbTrace(DB_CALL, (FTEXT(db_ctx, "(...)")));
    if (!context || !length)
	return (DWORD)ERROR_INVALID_PARAMETER;
    info_name_ptr.length = *length;
    info_name_ptr.text = name;
    code = supsys_call(context, READER_FUN_USER_DEFINED_NAME_GET, &info_name_ptr);
    if (code)
	return code;
    if (info_name_ptr.length == 0)
	return (DWORD)ERROR_NOT_SUPPORTED;
    *length = info_name_ptr.length;
    return code;
}

